import copy
import warnings

from . import _consts
from ._ci_netcdfattrs import NetCDFglobalattrs
from ._tables import mjson, xjson
from ._utils import (
    check_existance,
    get_time_range_as_str,
    kwargs_to_self,
    object_attrs_to_self,
)


class PostProcessing:
    """Class for post-processing ``index_calculator.processing`` object.

    Parameters
    ----------
    proc_obj: index_calculator.processing
        ``index_calculator.processing`` object
    institution_id: str (default: "N/A"), optional
        Short name of the institution calculating the climate indicator.
    institution: str (default: "N/A"), optional
        Long name of the institution calculating the climate indicator.
    contact: str (default: "N/A"), optional
        Mail contact of the institution calculating the climate indicator.

    Example
    -------
    Calculate a climate indicator `TG` from netcdf file on disk
    and do some post-processing::

        from pyhomogenize import open_xrdataset
        from index_calculator import preprocessing
        from index_calculator import processing
        from index_calculator import postprocessing

        netcdf_file = "tas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_"
                      "GERICS-REMO2015_v1_day_20010101-20051231.nc"
        ds = open_xrdataset(netcdf_file)

        preproc = preprocessing(ds)
        proc = processing(index="TG", preproc_obj=preproc)
        postproc = postprocessing(
            project="CORDEX",
            proc_obj=proc,
            institution="Helmholtz-Zentrum hereon GmbH,"
                        "Climate Service Center Germany",
            institution_id="GERICS",
            contact="gerics-cordex@hereon.de",
        )

        postproc_ds = postproc.postproc
    """

    def __init__(
        self,
        proc_obj=None,
        institution_id="N/A",
        institution="N/A",
        contact="N/A",
        split=True,
        **kwargs,
    ):
        if proc_obj is None:
            raise ValueError(
                "Select an index_calculator.Processing object. 'proc_obj=...'"
            )
        object_attrs_to_self(proc_obj, self)

        self.institution_id = check_existance(
            {"institution_id": institution_id},
            self,
        )
        self.institution = check_existance(
            {"institution": institution},
            self,
        )
        self.contact = check_existance({"contact": contact}, self)
        self.period = check_existance({"period": False}, self)
        self.split = check_existance({"split": split}, self)
        kwargs_to_self(kwargs, self)
        self.postproc = self._postprocessing()

    def _get_time_borders(self, times):
        left, right = get_time_range_as_str(times, self.fmt)
        return f"{left}-{right}"

    def _postprocessing(self):
        _mjson = copy.deepcopy(mjson)
        _xjson = copy.deepcopy(xjson)

        def adjust_attributes(dictionary, value):
            output = {}
            for key in dictionary.keys():
                if isinstance(dictionary[key], dict):
                    output[key] = adjust_attributes(dictionary[key], value)
                else:
                    output[key] = dictionary[key].format(**value)
            return output

        json = {}
        json[self.IDXname] = _mjson[self.IDXname]
        json["global_att"] = _xjson["global_att"]
        try:
            json["global_att"].update(_xjson[self.project]["global_att"])
        except KeyError:
            warnings.warn(f"Project {self.project} not known.")
        json = adjust_attributes(json, self.replacement)
        output = NetCDFglobalattrs(
            self,
            self.proc,
            json["global_att"],
        ).output
        for attr_name, attr_value in json[self.IDXname].items():
            output[self.CIname].attrs[attr_name] = attr_value
        associated_files = []
        for var_name in self.var_name:
            if "associated_files" not in self.ds[var_name].attrs.keys():
                continue
            assoc_files = self.ds[var_name].attrs["associated_files"]
            if isinstance(assoc_files, str):
                assoc_files = [assoc_files]
            for assoc_file in assoc_files:
                associated_files += [assoc_file]
        associated_files = ", ".join(associated_files)
        output[self.CIname].attrs["associated_files"] = associated_files
        if self.split is True:
            self.split = _consts._split[self.freq]
        trange = "ci_timerange_index"
        if self.split is False:
            output.attrs[trange] = self._get_time_borders(output.time.values)
            return output
        elif "time" in output.dims:
            olist = []
            for name, ds in output.resample({"time": self.split}):
                ds_ = ds.copy()
                ds_.attrs[trange] = self._get_time_borders(ds.time.values)
                olist.append(ds_)
            return olist
        elif "time" in output.coords:
            time_values = [
                output.time.values[()],
                output.time.values[()],
            ]
            output.attrs[trange] = self._get_time_borders(time_values)
            return output
        if hasattr(output, "ci_reference_period"):
            output.attrs[trange] = output.attrs["ci_reference_period"]
        return output



import warnings
from datetime import datetime as dt

import xclim as xc


def _get_version():
    __version__ = "unknown"
    try:
        from ._version import __version__
    except ImportError:
        pass
    return __version__


__version__ = _get_version()


class NetCDFattrs:
    """Class for writing key-value pairs to netCDF attributes."""

    def vanish_dictionary(self, dictionary):
        """Delete non-default key-value pairs from dictionary.

        Default key-value pairs are marked with '<'
        at the beginning of the value.

        Parameters
        ----------
        dictionary :  dict

        Returns
        -------
        dict,
            Dictionary with deleted non-default key-value pairs.
        """
        new_dictionary = {}
        for key, value in dictionary.items():
            if not value:
                continue
            if value[0] == "<":
                continue
            new_dictionary[key] = value
        return new_dictionary

    def write_dictionary_to_netcdf(self, database, dictionary):
        """Write key-value pairs to netCDF attributes.

        Parameters
        ----------
        database : xarray.Dataset or xarray.DataArray
            The key-value pairs will be written to the `database` attributes.
        dictionary: dict
            Dictionary with key-value pairs.

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            Database with new attributes
        """
        for k, v in dictionary.items():
            database.attrs[k] = v
        return database


class NetCDFglobalattrs(NetCDFattrs):
    """Class for writing key-value pairs to xarray.Dataset attributes."""

    def __init__(self, postproc, ds, dictionary={}):
        """Write key-value-pairs to xarray.Dataset attributes.

        Parameters
        ----------
        postproc: PostProcessing
            index_calculator PostProcessing object
        ds: xarray.Dataset
            The key-value pairs will be written to the `ds` attributes
        dictionary: dict, optional
            Dictionary with key-value pairs.

        Returns
        -------
        xarray.Dataset
            Database with new attributes
        """
        dictionary = self._create_dictionary(
            list_of_attributes=[
                "ci_additional_reference_period",
                "ci_creation_date",
                "ci_institute_id",
                "ci_institution",
                "ci_contact",
                "ci_name",
                "ci_reference_period",
                "ci_frequency",
                "ci_timerange_source",
                "ci_package_name",
                "ci_package_reference",
            ],
            obj=postproc,
        )
        dictionary = self.vanish_dictionary(dictionary)
        self.output = self.write_dictionary_to_netcdf(ds, dictionary)

    def _create_dictionary(self, list_of_attributes, obj):
        """Create attributes dictionary."""
        dictionary = {}
        for attribute in list_of_attributes:
            try:
                dictionary[attribute] = getattr(self, "_" + attribute)(obj)
            except AttributeError:
                warnings.warn(f"Could not set attribute {attribute}")
        return dictionary

    def _ci_additional_reference_period(self, obj):
        """Add additional reference period to dictionary."""
        return obj.period

    def _ci_creation_date(self, obj):
        """Add creation date to dictionary."""
        return dt.now().strftime("%Y-%m-%dT%H:%M:%SZ")

    def _ci_institution(self, obj):
        """Add institution long name to dictionary."""
        return obj.institution

    def _ci_institute_id(self, obj):
        """Add institution short name to dictionary."""
        return obj.institution_id

    def _ci_contact(self, obj):
        """Add contact to dictionary."""
        return obj.contact

    def _ci_name(self, obj):
        """Add climate index name to dictionary."""
        return obj.CIname

    def _ci_package_reference(self, obj):
        """Add index_calculator version to dictionary."""
        return f"xcalc_{__version__}"

    def _ci_package_name(self, obj):
        """Add xclim version to dictionary."""
        return f"xclim_{xc.__version__}"

    def _ci_reference_period(self, obj):
        """Add reference period to dictionary."""
        start = obj.base_period_time_range[0].replace("-", "")
        end = obj.base_period_time_range[1].replace("-", "")
        return f"{start}-{end}"

    def _ci_frequency(self, obj):
        """Add frequency to dictionary."""
        return obj.freq

    def _ci_timerange_source(self, obj):
        """Add time range of the source to dictionary."""
        return f"{obj.ATimeRange[0]}-{obj.ATimeRange[1]}"



#
# index_calculator documentation build configuration file
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys
import warnings
from subprocess import call

sys.path.insert(0, os.path.abspath(".."))

import index_calculator  # noqa

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.extlinks",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    #    "sphinxcontrib.mockautodoc",
    "numpydoc",
    "nbsphinx",
    "nbsphinx_link",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]


extlinks = {
    "issue": ("https://github.com/ludwiglierhammer/index_calculator/issues/%s", "#%s"),
    "pull": ("https://github.com/ludwiglierhammer/index_calculator/pull/%s", "#%s"),
}

autosummary_generate = True

napoleon_use_param = False
napoleon_use_rtype = False

numpydoc_class_members_toctree = True
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "index_calculator"
copyright = "2021-2022, Ludwig Lierhammer"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = index_calculator.__version__  # .split("+")[0]
# The full version, including alpha/beta/rc tags.
release = index_calculator.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "**.ipynb_checkpoints"]  # , "logo"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# does not work on rdt
# html_theme_options = {'logo_only': True}


# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = "logo/logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = "logo/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "index_calculatordoc"

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).

# If true, show URL addresses after external links.
# man_show_urls = False

# list conda packages
call("conda list", shell=True)

# disable warnings
warnings.filterwarnings("ignore")

notebooks = ()
print("\nBuilding notebooks:")
for nb in notebooks:
    # only render notebooks if necessary
    f1 = os.path.getmtime(nb + ".ipynb")
    try:
        f2 = os.path.getmtime(nb + ".rst")

        if f2 > f1:
            print(f" --- skipping: {nb}")
            continue
    except FileNotFoundError:
        pass

    call(
        (
            "jupyter nbconvert"
            " --to rst"
            " --template-file notebooks/tutorial_rst.tpl"
            " --ExecutePreprocessor.timeout=60"
            " --execute " + nb
        ),
        shell=True,
    )



"""Main module."""

from ._outputwriter import OutputWriter as outputwriter
from ._postprocessing import PostProcessing as postprocessing
from ._preprocessing import PreProcessing as preprocessing
from ._processing import Processing as processing
from ._utils import kwargs_to_self, object_attrs_to_self


class IndexCalculator:
    """Class for calculating climate indices with xclim.

    Parameters
    ----------
    write: bool (default: False), optional
        If True write climate index dataset on disk.

    Notes
    -----
    For more parameter information see:

        :func:`~index_calculator.preprocessing`
        :func:`~index_calculator.processing`
        :func:`~index_calculator.postprocessing`
        :func:`~index_calculator.outputwriter`

    Example
    -------
    Calculate climate indicator and write dataset as netcdf fiel on disk::

        from pyhomogenize import open_xrdataset
        from index_calculator import index_calculator

        netcdf_file = "tas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_"
                      "GERICS-REMO2015_v1_day_20010101-20051231.nc"
        ds = open_xrdataset(netcdf_file)

        idx = index_calculator(
                write=True,
                ds=ds,
                index="TG",
                project="CORDEX",
                institution_id="GERICS",
                institution="Helmholtz-Zentrum hereon GmbH,"
                            "Climate Service Center Germany",
                contact="gerics-cordex@hereon.de",
        )

        --> File written: cordex/climdex/EUR-11/GERICS/GERICS/MPI-M-MPI-ESM-LR/
                          historical/r3i1p1/GERICS-REMO2015/v1/year/TG/
                          TG_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_
                          GERICS-REMO2015_v1_day_GERICS_year_2001-2005.nc

    """

    def __init__(self, write=False, **kwargs):
        kwargs_to_self(kwargs, self)
        postproc_obj = self._compute(write=write, **kwargs)
        object_attrs_to_self(postproc_obj, self)

    def _compute(self, write=False, **kwargs):
        """Compute climate index."""
        preproc_obj = preprocessing(**kwargs)
        proc_obj = processing(preproc_obj=preproc_obj)
        postproc_obj = postprocessing(proc_obj=proc_obj)
        if write is True:
            outputwriter(
                postproc_obj=postproc_obj,
                **kwargs,
            )
        return postproc_obj



import pandas as pd
import pooch
import xarray as xr

url_base = (
    "https://github.com/climate-service-center/index_calculator_test_data/raw/main/"
)


def _pooch_retrieve(url, known_hash=None):
    return pooch.retrieve(url, known_hash=known_hash)


def tas_day_netcdf():
    url = (
        url_base
        + "/tas/day/tas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_GERICS-REMO2015_v1_day_20010101-20010107.nc"  # noqa
    )
    return _pooch_retrieve(url)


def tas_1hr_netcdf():
    url1 = (
        url_base
        + "/tas/1hr/tas_EUR-11_MPI-M-MPI-ESM-LR_rcp85_r3i1p1_GERICS-REMO2015_v1_1hr_200701010000-200701012300.nc"  # noqa
    )
    url2 = (
        url_base
        + "/tas/1hr/tas_EUR-11_MPI-M-MPI-ESM-LR_rcp85_r3i1p1_GERICS-REMO2015_v1_1hr_200701020000-200701022300.nc"  # noqa
    )
    url3 = (
        url_base
        + "/tas/1hr/tas_EUR-11_MPI-M-MPI-ESM-LR_rcp85_r3i1p1_GERICS-REMO2015_v1_1hr_200701030000-200701032300.nc"  # noqa
    )
    url4 = (
        url_base
        + "/tas/1hr/tas_EUR-11_MPI-M-MPI-ESM-LR_rcp85_r3i1p1_GERICS-REMO2015_v1_1hr_200701040000-200701042300.nc"  # noqa
    )
    url5 = (
        url_base
        + "/tas/1hr/tas_EUR-11_MPI-M-MPI-ESM-LR_rcp85_r3i1p1_GERICS-REMO2015_v1_1hr_200701050000-200701052300.nc"  # noqa
    )
    url6 = (
        url_base
        + "/tas/1hr/tas_EUR-11_MPI-M-MPI-ESM-LR_rcp85_r3i1p1_GERICS-REMO2015_v1_1hr_200701060000-200701062300.nc"  # noqa
    )
    url7 = (
        url_base
        + "/tas/1hr/tas_EUR-11_MPI-M-MPI-ESM-LR_rcp85_r3i1p1_GERICS-REMO2015_v1_1hr_200701070000-200701072300.nc"  # noqa
    )
    return [
        _pooch_retrieve(url)
        for url in [
            url1,
            url2,
            url3,
            url4,
            url5,
            url6,
            url7,
        ]
    ]


def tas_eobs_day_netcdf():
    url = (
        url_base + "/tas/day/tg_ens_mean_0.1deg_reg_v27.0e_20010101-20010107.nc"  # noqa
    )
    return _pooch_retrieve(url)


def pr_day_netcdf():
    url = (
        url_base
        + "/pr/day/pr_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_GERICS-REMO2015_v1_day_20010101-20010107.nc"  # noqa
    )
    return _pooch_retrieve(url)


def snw_day_netcdf():
    url = (
        url_base
        + "/snw/day/snw_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_GERICS-REMO2015_v1_day_20010101-20010107.nc"  # noqa
    )
    return _pooch_retrieve(url)


def uas_day_netcdf():
    url = (
        url_base
        + "/uas/day/uas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_GERICS-REMO2015_v1_day_20010101-20010107.nc"  # noqa
    )
    return _pooch_retrieve(url)


def vas_day_netcdf():
    url = (
        url_base
        + "/vas/day/vas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_GERICS-REMO2015_v1_day_20010101-20010107.nc"  # noqa
    )
    return _pooch_retrieve(url)


def tas_series(values, **kwargs):
    def _tas_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="tas",
            attrs={
                "standard_name": "air_temperature",
                "cell_methods": "time: mean within days",
                "units": "K",
            },
        )

    return _tas_series(values, **kwargs)


def pr_series(values, **kwargs):
    def _pr_series(values, start="1/1/2000", units="kg m-2 s-1", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="pr",
            attrs={
                "standard_name": "precipitation_flux",
                "cell_methods": "time: mean within days",
                "units": units,
            },
        )

    return _pr_series(values, **kwargs)


def tasmin_series(values, **kwargs):
    def _tasmin_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="tasmin",
            attrs={
                "standard_name": "air_temperature",
                "cell_methods": "time: minimum within days",
                "units": "K",
            },
        )

    return _tasmin_series(values, **kwargs)


def tasmax_series(values, **kwargs):
    def _tasmax_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="tasmax",
            attrs={
                "standard_name": "air_temperature",
                "cell_methods": "time: maximum within days",
                "units": "K",
            },
        )

    return _tasmax_series(values, **kwargs)


def prsn_series(values, **kwargs):
    def _prsn_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="prsn",
            attrs={
                "standard_name": "snowfall_flux",
                "cell_methods": "time: mean",
                "units": "kg m-2 s-1",
            },
        )

    return _prsn_series(values, **kwargs)


def snd_series(values, **kwargs):
    def _snd_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="snd",
            attrs={
                "standard_name": "surface_snow_thickness",
                "cell_methods": "time: mean",
                "units": "m",
            },
        )

    return _snd_series(values, **kwargs)


def snw_series(values, **kwargs):
    def _snw_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="snw",
            attrs={
                "standard_name": "surface_snow_amount",
                "cell_methods": "time: mean",
                "units": "kg m-2",
            },
        )

    return _snw_series(values, **kwargs)


def hurs_series(values, **kwargs):
    def _hurs_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="hurs",
            attrs={
                "standard_name": "relative_humidity",
                "cell_methods": "time: mean",
                "units": "%",
            },
        )

    return _hurs_series(values, **kwargs)


def rsds_series(values, **kwargs):
    def _rsds_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="rsds",
            attrs={
                "standard_name": "surface_downwelling_shortwave_flux",
                "cell_methods": "time: mean",
                "units": "W m-2",
            },
        )

    return _rsds_series(values, **kwargs)


def rsus_series(values, **kwargs):
    def _rsus_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="rsus",
            attrs={
                "standard_name": "surface_upwelling_shortwave_flux",
                "cell_methods": "time: mean",
                "units": "W m-2",
            },
        )

    return _rsus_series(values, **kwargs)


def rlds_series(values, **kwargs):
    def _rlds_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="rlds",
            attrs={
                "standard_name": "surface_downwelling_longwave_flux",
                "cell_methods": "time: mean",
                "units": "W m-2",
            },
        )

    return _rlds_series(values, **kwargs)


def rlus_series(values, **kwargs):
    def _rlus_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="rlus",
            attrs={
                "standard_name": "surface_upwelling_longwave_flux",
                "cell_methods": "time: mean",
                "units": "W m-2",
            },
        )

    return _rlus_series(values, **kwargs)


def sfcWind_series(values, **kwargs):
    def _sfcWind_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="sfcWind",
            attrs={
                "standard_name": "wind_speed",
                "cell_methods": "time: mean",
                "units": "m s-1",
            },
        )

    return _sfcWind_series(values, **kwargs)


def sfcWindmax_series(values, **kwargs):
    def _sfcWindmax_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="sfcWindmax",
            attrs={
                "standard_name": "wind_speed",
                "cell_methods": "time: max",
                "units": "m s-1",
            },
        )

    return _sfcWindmax_series(values, **kwargs)


def mrt_series(values, **kwargs):
    def _mrt_series(values, start="1/1/2000", freq="D"):
        coords = pd.date_range(start, periods=len(values), freq=freq)
        return xr.DataArray(
            values,
            coords=[coords],
            dims="time",
            name="mrt",
            attrs={
                "standard_name": "mean_radiant_temperature",
                "cell_methods": "time: mean",
                "units": "K",
            },
        )

    return _mrt_series(values, **kwargs)



import pytest  # noqa
from pyhomogenize import open_xrdataset

import index_calculator as xcalc

from .conftest import (
    pr_day_netcdf,
    snw_day_netcdf,
    tas_1hr_netcdf,
    tas_day_netcdf,
    tas_eobs_day_netcdf,
    uas_day_netcdf,
    vas_day_netcdf,
)


def test_processing():
    data = tas_day_netcdf()
    tas_ds = open_xrdataset(data)
    preproc = xcalc.preprocessing(
        tas_ds,
        freq="week",
        crop_time_axis=False,
        project="CORDEX",
    )
    proc = xcalc.processing("TG", preproc_obj=preproc)
    postproc = xcalc.postprocessing(
        proc_obj=proc,
        institution="test institution",
        institution_id="TEST",
        contact="test@test.de",
    )
    xcalc.outputwriter(postproc_obj=postproc)


def test_time_range_index_calculator():
    data = tas_day_netcdf()
    tas_ds = open_xrdataset(data)
    xcalc.index_calculator(
        ds=tas_ds,
        freq="week",
        index="TG",
        crop_time_axis=False,
        project="CORDEX",
        institution="test institution",
        institution_id="TEST",
        contact="test@test.de",
        write=True,
        time_range=["2001-01-01", "2001-01-07"],
    )


def test_thresh_index_calculator():
    data = pr_day_netcdf()
    pr_ds = open_xrdataset(data)
    xcalc.index_calculator(
        ds=pr_ds,
        freq="week",
        index="RX3day",
        crop_time_axis=False,
        project="CORDEX",
        institution_id="TEST",
    )
    xcalc.index_calculator(
        ds=pr_ds,
        freq="week",
        index="RXYYday",
        thresh=3,
        crop_time_axis=False,
        project="CORDEX",
        institution_id="TEST",
    )
    xcalc.index_calculator(
        ds=pr_ds,
        freq="week",
        index="RXYYday",
        crop_time_axis=False,
        project="CORDEX",
        institution_id="TEST",
    )


def test_perc_index_calculator():
    data = pr_day_netcdf()
    pr_ds = open_xrdataset(data)
    xcalc.index_calculator(
        ds=pr_ds,
        freq="week",
        index="RR95p",
        crop_time_axis=False,
        project="CORDEX",
        institution_id="TEST",
        base_period_time_range=["2001-01-01", "2001-01-07"],
    )
    xcalc.index_calculator(
        ds=pr_ds,
        freq="week",
        index="R95p",
        crop_time_axis=False,
        project="CORDEX",
        institution_id="TEST",
        base_period_time_range=["2001-01-01", "2001-01-07"],
    )


def test_snw_index_calculator():
    data = snw_day_netcdf()
    snw_ds = open_xrdataset(data)
    xcalc.index_calculator(
        ds=snw_ds,
        freq="week",
        index="SCD",
        crop_time_axis=False,
        project="CORDEX",
        institution="test institution",
        institution_id="TEST",
        contact="test@test.de",
        write=True,
    )


def test_snw_const_index_calculator():
    data = snw_day_netcdf()
    snw_ds = open_xrdataset(data)
    xcalc.index_calculator(
        ds=snw_ds,
        freq="week",
        index="SCD",
        crop_time_axis=False,
        project="CORDEX",
        institution="test institution",
        institution_id="TEST",
        contact="test@test.de",
        write=True,
        const="300 kg m-3",
    )


def test_uas_vas_index_calculator():
    data_uas = uas_day_netcdf()
    data_vas = vas_day_netcdf()
    sfcWind_ds = open_xrdataset([data_uas, data_vas])
    xcalc.index_calculator(
        ds=sfcWind_ds,
        freq="week",
        index="FG",
        crop_time_axis=False,
        project="CORDEX",
        institution="test institution",
        institution_id="TEST",
        contact="test@test.de",
        write=True,
    )


def test_1hr_index_calculator():
    data = tas_1hr_netcdf()
    tas_ds = open_xrdataset(data)
    xcalc.index_calculator(
        ds=tas_ds,
        freq="week",
        index="TG",
        crop_time_axis=False,
        project="CORDEX",
        institution="test institution",
        institution_id="TEST",
        contact="test@test.de",
        write=True,
    )


def test_eobs_index_calculator():
    data = tas_eobs_day_netcdf()
    tas_ds = open_xrdataset(data)
    xcalc.index_calculator(
        ds=tas_ds,
        freq="week",
        index="TG",
        crop_time_axis=False,
        project="EOBS",
        institution="test institution",
        institution_id="TEST",
        contact="test@test.de",
        write=True,
    )



import re
import warnings

from pyhomogenize import basics


def object_attrs_to_self(obj, slf, overwrite=True):
    """Copy object attributes to new object."""
    for attr in dir(obj):
        if attr[0] == "_":
            continue
        if overwrite is False and hasattr(slf, attr):
            continue
        if attr[0].isalpha():
            setattr(slf, attr, getattr(obj, attr))


def kwargs_to_self(kwargs, slf):
    """Write kwargs to new object."""
    if not hasattr(slf, "kwargs"):
        setattr(slf, "kwargs", kwargs)
    else:
        slf.kwargs.update(kwargs)
    for key, value in kwargs.items():
        if not hasattr(slf, key):
            setattr(slf, key, value)
        if getattr(slf, key) is None:
            setattr(slf, key, value)


def check_existance(attr_dict, slf):
    """Check existance of values."""
    for key, value in attr_dict.items():
        test = False
        if value is None:
            method = "raise"
            test = True
        if value == "N/A":
            method = "warn"
            test = True
        if value is True:
            method = ""
            test = True
        if test:
            if hasattr(slf, key):
                return getattr(slf, key)
            else:
                msg = f"No {key} is selected. '{key}=...'"
                if method == "raise":
                    raise ValueError(msg)
                elif method == "warn":
                    warnings.warn(msg)
        return value


def get_time_range_as_str(time, fmt):
    ts = time[0]
    te = time[-1]
    if not isinstance(ts, str):
        ts = basics().date_to_str(ts, fmt)
    if not isinstance(te, str):
        te = basics().date_to_str(te, fmt)
    return [ts, te]


def get_alpha_name(var_name):
    return "".join(filter(lambda x: x.isalpha(), var_name))


def get_numb_name(var_name):
    return "".join(filter(lambda x: x.isdigit(), var_name))


def get_replace_name(var_name):
    return re.sub(r"\d+", "YY", var_name)



from copy import deepcopy
from inspect import signature
from warnings import warn

import dask
import xarray as xr
from xclim.core.calendar import percentile_doy
from xclim.core.units import convert_units_to

from ._consts import _base_period as BASE_PERIOD


class ClimateIndicator:
    def __init__(self):
        self.func = None
        self.units = {}
        self.date_bounds = None
        self.base_period_time_range = BASE_PERIOD
        self.split_large_chunks = True

    def _thresh_string(self, thresh, units):
        if isinstance(thresh, str):
            return thresh
        else:
            return f"{str(thresh)} {units}"

    def _get_da(self, dictionary, var):
        if "ds" in dictionary.keys():
            return dictionary["ds"][var]
        elif var in dictionary.keys():
            return dictionary[var]
        raise ValueError("Variable {} not found!")

    def _clean_up_params(self, params, func, exceptions=[]):
        del_list = []
        if not callable(func):
            return params
        for param in params.keys():
            if param in exceptions:
                continue
            if param not in signature(func).parameters.keys():
                warn(
                    f"Function {func} does not provide parameter {param}\n"
                    "The parameter will be deleted \n"
                    "from parameter list"
                )
                del_list.append(param)
        for param in del_list:
            del params[param]
        return params

    def _set_default_if_None(self, params):
        params_ = {}
        for k, v in params.items():
            if isinstance(v, dict):
                v = self._set_default_if_None(v)
            if v is None:
                if hasattr(self, k):
                    v = getattr(self, k)
                else:
                    continue
            params_[k] = v
        return params_

    def _add_units(self, params):
        params_ = {}
        for k, v in params.items():
            if isinstance(v, dict):
                v = self._add_units(v)
            if k in self.units.keys():
                v = self._thresh_string(v, self.units[k])
            params_[k] = v
        return params_

    def _dates_to_bounds(self, kwargs):
        kwargs["date_bounds"] = (kwargs["start_date"], kwargs["end_date"])
        return kwargs

    def _filter_out_small_values(self, da, thresh="1mm/day", context="hydro"):
        thresh = convert_units_to(thresh, da, context=context)
        return da.where(da > thresh)

    def _get_percentile(self, da, per, base_period_time_range):
        if isinstance(per, xr.Dataset):
            return per["per"]
        elif isinstance(per, xr.DataArray):
            return per
        tslice = slice(base_period_time_range[0], base_period_time_range[1])
        base_period = da.sel(time=tslice)
        with dask.config.set(**{"array.slicing.split_large_chunks": False}):
            per_doy = percentile_doy(base_period, per=per)
            per_doy_comp = per_doy.compute()
        return per_doy_comp.sel(percentiles=per)

    def _preprocessing(self, da, method=None, **kwargs):
        kwargs_ = deepcopy(kwargs)
        kwargs = self._clean_up_params(
            params=kwargs_, func=self._filter_out_small_values
        )
        if method == "precipitation":
            da = self._filter_out_small_values(da, **kwargs)
        return da

    def _get_percentiles(self, params, kwargs):
        self.split_large_chunks = False
        for name_per, kwargs_dict in kwargs["percentiles"].items():
            if name_per in kwargs.keys():
                if isinstance(kwargs[name_per], (xr.DataArray, xr.Dataset)):
                    continue
            da = self._get_da(params, kwargs_dict["variable"])
            da = self._preprocessing(
                da,
                method=kwargs_dict["method"],
                **kwargs,
            )
            kwargs[name_per] = self._get_percentile(
                da, kwargs_dict["per"], kwargs["base_period_time_range"]
            )
        return kwargs

    def compute_climate_indicator(self, params, **kwargs):
        params = self._clean_up_params(params=params, func=self.func)
        kwargs = self._set_default_if_None(kwargs)
        kwargs = self._add_units(kwargs)
        if "percentiles" in kwargs.keys():
            kwargs = self._get_percentiles(params, kwargs)
        if self.date_bounds is True:
            kwargs = self._dates_to_bounds(kwargs)
        if self.func is None:
            return kwargs
        kwargs = self._clean_up_params(
            params=kwargs, func=self.func, exceptions=["date_bounds"]
        )
        if self.split_large_chunks is True:
            return self.func(**params, **kwargs)
        if self.split_large_chunks is False:
            with dask.config.set(
                **{
                    "array.slicing.split_large_chunks": False,
                }
            ):
                return self.func(**params, **kwargs)



import inspect

import xclim as xc


class ConvertVariables:
    def __init__(self, ds, **params):
        self.ds = ds
        self.params = params

    def _params(self, func, params):
        fparams = inspect.signature(func).parameters
        return {k: v for k, v in params.items() if k in fparams.keys()}

    def sfcWind(self):
        return xc.indicators.convert.wind_speed_from_vector(
            ds=self.ds,
        )[0]

    def snd(self):
        params = self._params(xc.land.snw_to_snd, self.params)
        return xc.land.snw_to_snd(ds=self.ds, **params)



import xclim as xc

from ._climate_indicator import ClimateIndicator


class CD(ClimateIndicator):
    """Number of cold and dry days (tas, pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.cold_and_dry_days

    def compute(
        self,
        base_period_time_range=None,
        tas_per=None,
        pr_per=None,
        **params,
    ):
        """Calculate number of cold and dry days.

        Parameters
        ----------
        tas_per: xr.DataArray, optional
            Mean temperature 25th percentile reference value below which
            a day is considered as a cold day.
        pr_per: xr.DataArray, optional
            Precipitation 25th percentile reference value below which
            a day is considered as a dry day.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tas_per` and/or `pr_per`.
            This will be used only if `tas_per` and/or `pr_per` is None.

        Returns
        -------
        xarray.DataArray:
            Number of days where cold and dry conditions coincide.
            If temperature is below {tas_per} a day is considered as a
            cold day.
            If precipitation is below {pr_per} a day is considered as a
            dry day.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.cold_and_dry_days
        """
        percentiles = {
            "tas_per": {
                "variable": "tas",
                "per": 25,
                "method": "temperature",
            },
            "pr_per": {
                "variable": "pr",
                "per": 25,
                "method": "precipitation",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tas_per=tas_per,
            pr_per=pr_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class CDD(ClimateIndicator):
    """Maximum consecutive dry days (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.maximum_consecutive_dry_days

    def compute(self, thresh=None, **params):
        """Calculate maximum consecutive dry days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation below which a day is considered
            as a dry day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Maximum consecutive dry days (precipitation < {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.maximum_consecutive_dry_days
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class CFD(ClimateIndicator):
    """Maximum number of consecutive frost days (tasmin)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.consecutive_frost_days

    def compute(self, **params):
        """Calculate maximum number of consecutive frost days.

        Returns
        -------
        xarray.DataArray
            Maximum number of consecutive frost days (tasmin < 0 degC).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.consecutive_frost_days
        """
        return self.compute_climate_indicator(params=params)


class CHDYYx(ClimateIndicator):
    """Maximum number of consecutive heat days (tasmax)."""

    def __init__(self):
        super().__init__()
        self.thresh = 30
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.maximum_consecutive_warm_days

    def compute(self, thresh=None, **params):
        """Calculate maximum number of consecutive heat days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold maximum temperature above which a day is considered
            as a heat day (default: 30 degC).
            If type of threshold is an integer the unit is set to degC.

        Returns
        -------
        xarray.DataArray
            Maximum number of consecutive heat days (tasmax > {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.maximum_consecutive_warm_days
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class CSDI(ClimateIndicator):
    """Cold spell duration index (tasmin)."""

    def __init__(self):
        super().__init__()
        self.window = 6
        self.func = xc.atmos.cold_spell_duration_index

    def compute(
        self,
        window=None,
        base_period_time_range=None,
        tasmin_per=None,
        **params,
    ):
        """Calculate cold spell duration index.

        Parameters
        ----------
        window: int, optional
            Minimum number of days with temperature below `tasmin_per`
            to qualify as a cold spell (default: 6).
        tasmin_per: xr.DataArray, optional
            Minimum temperature 10th percentile reference value.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tasmin_per`.
            This will be used only if `tasmin_per` is None.

        Returns
        -------
        xarray.DataArray
            Number of days part of a 10th percentile cold spell.
            At least {window} consecutive days.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.cold_spell_duration_index
        """
        percentiles = {
            "tasmin_per": {
                "variable": "tasmin",
                "per": 10,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            window=window,
            tasmin_per=tasmin_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class CSU(ClimateIndicator):
    """Maximum consecutive summer days (tasmax)."""

    def __init__(self):
        super().__init__()
        self.thresh = 25
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.maximum_consecutive_warm_days

    def compute(self, thresh=None, **params):
        """Calculate maximum consecutive summer days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold  maximum temperature above which a day is considered
            as a summer day (default: 25 degC).
            If type of threshold is an integer the unit is set to degC.

        Returns
        -------
        xarray.DataArray
            Maximum consecutive summer days (tasmax > {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.maximum_consecutive_warm_days
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class CW(ClimateIndicator):
    """Number of cold and wet days (tas, pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.cold_and_wet_days

    def compute(
        self,
        base_period_time_range=None,
        tas_per=None,
        pr_per=None,
        **params,
    ):
        """Calculate number of cold and wet days.

        Parameters
        ----------
        tas_per: xr.DataArray, optional
            Mean temperature 25th percentile reference value below which
            a day is considered as a cold day.
        pr_per: xr.DataArray, optional
            Precipitation 75th percentile reference value above which
            a day is considered as a wet day.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tas_per` and/or `pr_per`.
            This will be used only if `tas_per` and/or `pr_per` is None.

        Returns
        -------
        xarray.DataArray:
            Number of days where cold and wet conditions coincide.
            If temperature is below {tas_per} a day is considered as a
            cold day.
            If precipitation is above {pr_per} a day is considered as a
            wet day.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.cold_and_wet_days
        """
        percentiles = {
            "tas_per": {
                "variable": "tas",
                "per": 25,
                "method": "temperature",
            },
            "pr_per": {
                "variable": "pr",
                "per": 75,
                "method": "precipitation",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tas_per=tas_per,
            pr_per=pr_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class CWD(ClimateIndicator):
    """Consecutive wet days (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.maximum_consecutive_wet_days

    def compute(self, thresh=None, **params):
        """Calculate maximum consecutive wet days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation at or above which a day is considered
            as a wet day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Maximum consecutive wet days (pr >= {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.maximum_consecutive_wet_days
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class DD(ClimateIndicator):
    """Number of dry days (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.dry_days

    def compute(self, thresh=None, **params):
        """Calculate number of dry days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation below which a day is considered
            as a dry day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Number of dry days (pr < {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.dry_days
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class DSf(ClimateIndicator):
    """Number of dry spells (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.window = 5
        self.units = {"thresh": "mm"}
        self.func = xc.atmos.dry_spell_frequency

    def compute(self, thresh=None, window=None, **params):
        """Calculate number of dry spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation below which a day is considered
            as a dry day (default: 1 mm).
            If type of threshold is an integer the unit is set to mm.
        window: int, optional
            Minimum number of days with precipitation below threshold
            to qualify as a dry spell (default: 5).

        Returns
        -------
        xarray.DataArray
            Number of dry periods of minimum {window} days (pr < {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.dry_spell_frequency
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class DSx(ClimateIndicator):
    """Maximum length of dry spells (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.window = 1
        self.units = {"thresh": "mm"}
        self.func = xc.atmos.dry_spell_max_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate maximum length of dry spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation below which a day is considered
            as a dry day (default: 1 mm).
            If type of threshold is an integer the unit is set to mm.
        window: int, optional
            Minimum number of days with precipitation below threshold
            to qualify as a dry spell (default: 1).

        Returns
        -------
        xarray.DataArray
            Maximum length of dry spells of at least {window} consecutive days
            with precipitation below {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.dry_spell_max_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class DSn(ClimateIndicator):
    """Total number of days in dry spells (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.window = 5
        self.units = {"thresh": "mm"}
        self.func = xc.atmos.dry_spell_total_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate total number of days in dry spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation below which a day is considered
            as a dry day (default: 1 mm).
            If type of threshold is an integer the unit is set to mm.
        window: int, optional
            Minimum number of days with precipitation below threshold
            to qualify as a dry spell (default: 5).

        Returns
        -------
        xarray.DataArray
            Total number of days in dry spells of at least {window}
            consecutive days with precipitation below {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.dry_spell_total_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class WSf(ClimateIndicator):
    """Number of wet spells (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.window = 5
        self.units = {"thresh": "mm"}
        self.func = xc.atmos.wet_spell_frequency

    def compute(self, thresh=None, window=None, **params):
        """Calculate number of wet spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm).
            If type of threshold is an integer the unit is set to mm.
        window: int, optional
            Minimum number of days with precipitation above threshold
            to qualify as a wet spell (default: 5).

        Returns
        -------
        xarray.DataArray
            Number of wet periods of minimum {window} days (pr >= {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wet_spell_frequency
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class WSx(ClimateIndicator):
    """Maximum length of wet spells (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.window = 1
        self.units = {"thresh": "mm"}
        self.func = xc.atmos.wet_spell_max_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate maximum length of wet spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm).
            If type of threshold is an integer the unit is set to mm.
        window: int, optional
            Minimum number of days with precipitation above threshold
            to qualify as a wet spell (default: 1).

        Returns
        -------
        xarray.DataArray
            Maximum length of wet spells of at least {window} consecutive days
            with precipitation above {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wet_spell_max_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class WSn(ClimateIndicator):
    """Total number of days in wet spells (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.window = 5
        self.units = {"thresh": "mm"}
        self.func = xc.atmos.wet_spell_total_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate total number of days in wet spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm).
            If type of threshold is an integer the unit is set to mm.
        window: int, optional
            Minimum number of days with precipitation above threshold
            to qualify as a wet spell (default: 5).

        Returns
        -------
        xarray.DataArray
            Total numer of days in wet spells of at least {window}
            consecutive days with precipitation above {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wet_spell_total_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class DTR(ClimateIndicator):
    """Mean temperature rnage (tasmax, tasmin)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.daily_temperature_range

    def compute(self, **params):
        """Calculate mean of daily temperature range.

        Returns
        -------
        xarray.DataArray
            Mean of daily temperature range.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.daily_temperature_range
        """
        return self.compute_climate_indicator(params=params)


class FD(ClimateIndicator):
    """Number of frost days (tasmin)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.frost_days

    def compute(self, **params):
        """Calculate number of frost days (tasmin < 0.0 degC).

        Returns
        -------
        xarray.DataArray
            Number of frost days (tasmin < 0.0 degC).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.frost_days
        """
        return self.compute_climate_indicator(params=params)


class LFD(ClimateIndicator):
    """Number of late frost days (tasmin)."""

    def __init__(self):
        super().__init__()
        self.start_date = "04-01"
        self.end_date = "06-30"
        self.date_bounds = True
        self.func = xc.atmos.late_frost_days

    def compute(self, start_date=None, end_date=None, **params):
        """Calculate number of late frost days (tasmin < 0.0 degC).

        Parameters
        ----------
        start_date: str, optional
            Left bound of the period to be considered.
        end_date: str, optional
            Right bound of the period to be considered.

        Returns
        -------
        xarray.DataArray
            Number of frost days (tasmin < 0.0)
            between {start_date} and {end_date}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.late_frost_days
        """
        return self.compute_climate_indicator(
            params=params, start_date=start_date, end_date=end_date
        )


class ID(ClimateIndicator):
    """Number of ice days (tasmax)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.ice_days

    def compute(self, **params):
        """Calculate number of ice days (tasmax < 0.0 degC).

        Returns
        -------
        xarray.DataArray
            Number of ice days (tasmax < 0.0 degC).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.ice_days
        """
        return self.compute_climate_indicator(params=params)


class GD(ClimateIndicator):
    """Cumulative growing degree days (tas)."""

    def __init__(self):
        super().__init__()
        self.thresh = 4
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.growing_degree_days

    def compute(self, thresh=None, **params):
        """Calculate cumulative growing degree days (tas > thresh).

        Parameters
        ----------
        thresh: int or string, optional
            Threshold temperature above which the daily temperature will
            be added to the cumulative sum of temperature degrees.

        Returns
        -------
        xarray.DataArray
            Cumulative growing degree days (tas > {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.growing_degree_days
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class HD17(ClimateIndicator):
    """Cumulative heating degree days (tas)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.heating_degree_days

    def compute(self, **params):
        """Calculate cumulative heating degree days (tas < 17 degC).

        Returns
        -------
        xarray.DataArray
            Cumulative heating degree days (tas > 17 degC).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.heating_degree_days
        """
        return self.compute_climate_indicator(params=params)


class PRCPTOT(ClimateIndicator):
    """Total precipitation amount (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.wet_precip_accumulation

    def compute(self, thresh=None, **params):
        """Calculate total precipitation amount.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Total precipitation amount of wet days (precip > {thresh})

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wet_precip_accumulation
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class RR(ClimateIndicator):
    """Total precipitation (pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.precip_accumulation

    def compute(self, **params):
        """Calculate total precipitation.

        Returns
        -------
        xarray.DataArray
            Total precipitation.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.precip_accumulation
        """
        return self.compute_climate_indicator(params=params)


class RRm(ClimateIndicator):
    """Mean daily precipitation (pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.precip_average

    def compute(self, **params):
        """Calculate mean daily precipitation.

        Returns
        -------
        xarray.DataArray
            Mean daily precipitation.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.precip_average
        """
        return self.compute_climate_indicator(params=params)


class RR1(ClimateIndicator):
    """Number of wet days (pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.wetdays

    def compute(self, **params):
        """Calculate number of wet days (pr >= 1 mm/day).

        Returns
        -------
        xarray.DataArray
            Number of wet days (pr >= 1 mm/day).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wetdays
        """
        return self.compute_climate_indicator(params=params, thresh="1 mm/day")


class R10mm(ClimateIndicator):
    """Number of heavy precipitation days (pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.wetdays

    def compute(self, **params):
        """Calculate number of wet days (pr >= 10 mm/day).

        Returns
        -------
        xarray.DataArray
            Number of wet days (pr >= 10 mm/day).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wetdays
        """
        return self.compute_climate_indicator(
            params=params,
            thresh="10 mm/day",
        )


class R20mm(ClimateIndicator):
    """Number of very heavy precipitation days (pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.wetdays

    def compute(self, **params):
        """Calculate number of wet days (pr >= 20 mm/day).

        Returns
        -------
        xarray.DataArray
            Number of wet days (pr >= 20 mm/day).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wetdays
        """
        return self.compute_climate_indicator(
            params=params,
            thresh="20 mm/day",
        )


class R25mm(ClimateIndicator):
    """Number of super heavy precipitation days (pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.wetdays

    def compute(self, **params):
        """Calculate number of wet days (pr >= 25 mm/day).

        Returns
        -------
        xarray.DataArray
            Number of wet days (pr >= 25 mm/day).

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wetdays
        """
        return self.compute_climate_indicator(
            params=params,
            thresh="25 mm/day",
        )


class RRYYp(ClimateIndicator):
    """Precip percentil value for wet days (pr)."""

    def __init__(self):
        super().__init__()
        self.per = 75
        self.thresh = 1
        self.units = {"thresh": "mm/day"}

    def compute(
        self,
        per=None,
        thresh=None,
        base_period_time_range=None,
        **params,
    ):
        """Calculate precip percentile reference value
        for wet days (pr > thresh).

        Parameters
        ----------
        per: int, optional
            Percentile value.
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating the precip percentile
            reference value.

        Returns
        -------
        xarray.DataArray
            Precip {per}th percentil reference value
            for wet days (pr > {thresh}).
        """
        percentiles = {
            "pr_per": {
                "variable": "pr",
                "per": per,
                "method": "precipitation",
                "thresh": thresh,
            },
        }
        results = self.compute_climate_indicator(
            params=params,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )
        return results["pr_per"]


class RYYp(ClimateIndicator):
    """Number of wet days with precip over a given percentile (pr)."""

    def __init__(self):
        super().__init__()
        self.per = 75
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.days_over_precip_doy_thresh

    def compute(
        self,
        per=None,
        thresh=None,
        base_period_time_range=None,
        **params,
    ):
        """Calculate number of wet days (pr > thresh)
        with precip over a given percentile.

        Parameters
        ----------
        per: int, optional
            Precipitation percentile value.
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating the precip
            percentile reference value.

        Returns
        -------
        xarray.DataArray
            Number of wet days (pyr > {thresh}) over a {per}th percentile.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.days_over_precip_doy_thresh
        """
        percentiles = {
            "pr_per": {
                "variable": "pr",
                "per": per,
                "method": "precipitation",
                "thresh": thresh,
            },
        }
        return self.compute_climate_indicator(
            params=params,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class RYYmm(ClimateIndicator):
    """Number of days with precip over threshold (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 25
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.wetdays

    def compute(self, thresh=None, **params):
        """Calculate number of wet days.

        Parameters
        ----------
        thresh: int or str, optional
        Threshold precipitation above which a day is considered
            as a wet day (default: 25 mm/day).
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Number of wet days (pr > {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.wetdays
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class RX1day(ClimateIndicator):
    """Maximum 1-day total precipitation (pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.max_1day_precipitation_amount

    def compute(self, **params):
        """Calculate maximum 1-day total precipitation.

        Returns
        -------
        xarray.DataArray
            Maximum 1-day total precipitation.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.max_1day_precipitation_amount
        """
        return self.compute_climate_indicator(params=params)


class RXYYday(ClimateIndicator):
    """Maximum n-day total precipitation (pr)."""

    def __init__(self):
        super().__init__()
        self.window = 5
        self.func = xc.atmos.max_n_day_precipitation_amount

    def compute(self, window=None, **params):
        """Calculate maximum {window}-day total precipitation.

        Parameters
        ----------
        window: int, optional
            Window size in days (default: 5).

        Returns
        -------
        xarray.DataArray
            Maximum {window}-day total precipitation.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.max_n_day_precipitation_amount
        """
        return self.compute_climate_indicator(params=params, window=window)


class RYYpTOT(ClimateIndicator):
    """Precipitation fraction with precip above percentile on wet days (pr)"""

    def __init__(self):
        super().__init__()
        self.per = 75
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.fraction_over_precip_thresh

    def compute(
        self,
        per=None,
        thresh=None,
        base_period_time_range=None,
        **params,
    ):
        """Calculate precipitation fraction with precip above percentile
        on wet days (pr > thresh).

        Parameters
        ----------
        per: int, optional
            Percentile value.
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calulating the precip percentile
            reference value.

        Returns
        -------
        xarray.DataArray
            Precipitation fraction with precip > {perc}th percentile
            on wet days (pr > {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.fraction_over_precip_thresh
        """
        percentiles = {
            "pr_per": {
                "variable": "pr",
                "per": per,
                "method": "precipitation",
                "thresh": thresh,
            },
        }
        return self.compute_climate_indicator(
            params=params,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class SDII(ClimateIndicator):
    """Average precipitation during wet days (pr)."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.daily_pr_intensity

    def compute(self, thresh=None, **params):
        """Calculate average precipitation during wet days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold precipitation above which a day is considered
            as a wet day (default: 1 mm/day).
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Average precipitation during wet days.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.daily_pr_intensity
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class SU(ClimateIndicator):
    """Number of summer days (tasmax)."""

    def __init__(self):
        super().__init__()
        self.thresh = 25
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.tx_days_above

    def compute(self, thresh=None, **params):
        """Calculate number of summer days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold maximum temperature above which to day is considered
            as a summer day (default: 25 degC).
            If type of threshold is an integer the unit is set to degC.

        Returns
        -------
        xarray.DataArray
            Number of summer days (tx > {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tx_days_above
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class CMD(ClimateIndicator):
    """Number of calm days (sfcWind)."""

    def __init__(self):
        super().__init__()
        self.thresh = 2
        self.units = {"thresh": "m s-1"}
        self.func = xc.atmos.calm_days

    def compute(self, thresh=None, **params):
        """Calculate number of calm days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold wind speed below which a day is considered
            as a calm day (default: 2 m s-1).
            If type of threshold is an integer the unit is set to m s-1.

        Returns
        -------
        xarray.DataArray
            Number of calm days ( fg > {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.calm_days

        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class SQI(ClimateIndicator):
    """Number of uncomfortable sleep events (tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh = 18
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.tn_days_above

    def compute(self, thresh=None, **params):
        """Calculate number of uncomfortable sleep events.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold minimum temperature below which a day has
            a uncomfortable sleep event (default: 18 degC).
            If type of threshold is an integer the unit is set to degC.

        Returns
        -------
        xarray.DataArray
            Number of uncomfortable sleep events (tn < {thresh}).

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tn_days_above
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class TG(ClimateIndicator):
    """Mean mean temperature (tas)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tg_mean

    def compute(self, **params):
        """Calculate mean daily mean temperature.

        Returns
        -------
        xarray.DataArray
            Mean daily mean temperature.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tg_mean
        """
        return self.compute_climate_indicator(params=params)


class TG10p(ClimateIndicator):
    """Fraction of days with mean temperature < 10th percentile (tas)."""

    def __init__(self):
        super().__init__()
        self.tas_per = None
        self.func = xc.atmos.tg10p

    def compute(
        self,
        base_period_time_range=None,
        tas_per=None,
        **params,
    ):
        """Calculate fraction of days with mean temperature < 10th percentile.

        Parameters
        ----------
        tas_per: xr.DataArray, optional
            Temperature 10th percentile reference value.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound is
            end year string for calculating `tas_per`.
            This will be used only if `tas_per` is None.

        Returns
        -------
        xarray.DataArray
            Fraction of days with mean temperature < 10th percentile".

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tg10p
        """
        percentiles = {
            "tas_per": {
                "variable": "tas",
                "per": 10,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tas_per=tas_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class TG90p(ClimateIndicator):
    """Fraction of days with mean temperature > 90th percentile (tas)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tg90p

    def compute(
        self,
        base_period_time_range=None,
        tas_per=None,
        **params,
    ):
        """Calculate fraction of days with mean temperature > 90th percentile.

         Parameters
         ----------
         tas_per: xr.DataArray, optional
             Temperature 90th percentile reference value.
         base_period_time_range: list, optional
             List with left bound is start year string and right bound
             is end year string for calculating `tas_per`.
             This will be used only if `tas_per` is None.

        Returns
         -------
         xarray.DataArray
             Fraction of days with mean temperature > 90th percentile".

         Notes
         -----
         For more information on the input parameters see:
             https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tg90p
        """
        percentiles = {
            "tas_per": {
                "variable": "tas",
                "per": 90,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tas_per=tas_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class TR(ClimateIndicator):
    """Number of tropical nights (tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh = 20
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.tn_days_above

    def compute(self, thresh=None, **params):
        """Calculate number of tropical nights.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold minimum temperature on which a day
            has a tropical night (default: 20 degC).
            If type of threshold is an integer the unit is set to degC.

        Returns
        -------
        xarray.DataArray
            Number of tropical nights (tn > {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tn_days_above
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class TX(ClimateIndicator):
    """Mean maximum temperature (tasmax)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tx_mean

    def compute(self, **params):
        """Calculate mean daily maximum temperature.

        Returns
        -------
        xarray.DataArray
            Mean daily maximum temperature.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tx_mean
        """
        return self.compute_climate_indicator(params=params)


class TX10p(ClimateIndicator):
    """Fraction of days with max temperature < 10th percentile (tasmax)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tx10p

    def compute(
        self,
        base_period_time_range=None,
        tasmax_per=None,
        **params,
    ):
        """Calculate fraction of days with max temperature < 10th percentile.

        Parameters
        ----------
        tasmax_per: xr.DataArray, optional
            Maximum temperature 10th percentile reference value.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tasmax_per`.
            This will be used only if `tasmax_per` is None.

        Returns
        -------
        xarray.DataArray
            Fraction of days with maximum temperature < 10th percentile".

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tx10p
        """
        percentiles = {
            "tasmax_per": {
                "variable": "tasmax",
                "per": 10,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tasmax_per=tasmax_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class TX90p(ClimateIndicator):
    """Fraction of days with max temperature > 90th percentile (tasmax)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tx90p

    def compute(
        self,
        base_period_time_range=None,
        tasmax_per=None,
        **params,
    ):
        """Calculate fraction of days with max temperature > 90th percentile.

        Parameters
        ----------
        tasmax_per: xr.DataArray, optional
            Maximum temperature 90th percentile reference value.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tasmax_per`.
            This will be used only if `tasmax_per` is None.

        Returns
        -------
        xarray.DataArray
            Fraction of days with maximum temperature > 90th percentile".

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tx90p
        """
        percentiles = {
            "tasmax_per": {
                "variable": "tasmax",
                "per": 90,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tasmax_per=tasmax_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class TXn(ClimateIndicator):
    """Minimum maximum temperature (tasmax)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tx_min

    def compute(self, **params):
        """Calculate minimum daily maximum temperature.

        Returns
        -------
        xarray.DataArray
            Minimum daily maximum temperature.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tx_min
        """
        return self.compute_climate_indicator(params=params)


class TXx(ClimateIndicator):
    """Maximum maximum temperature (tasmax)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tx_max

    def compute(self, **params):
        """Calculate maximum daily maximum temperature.

        Returns
        -------
        xarray.DataArray
            Maximum daily maximum temperature.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tx_max
        """
        return self.compute_climate_indicator(params=params)


class TN(ClimateIndicator):
    """Mean minimum temperature (tasmin)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tn_mean

    def compute(self, **params):
        """Calculate mean daily minimum temperature.

        Returns
        -------
        xarray.DataArray
            Mean daily minimum temperature.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tn_mean
        """
        return self.compute_climate_indicator(params=params)


class TN10p(ClimateIndicator):
    """Fraction of days with min temperature < 10th percentile."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tn10p

    def compute(
        self,
        base_period_time_range=None,
        tasmin_per=None,
        **params,
    ):
        """Calculate fraction of days with min temperature < 10th percentile.

        Parameters
        ----------
        tasmin_per: xr.DataArray, optional
            Minimum temperature 10th percentile reference value.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tasmin_per`.
            This will be used only if `tasmin_per` is None.

        Returns
        -------
        xarray.DataArray
            Fraction of days with minimum temperature < 10th percentile".

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tn10p
        """
        percentiles = {
            "tasmin_per": {
                "variable": "tasmin",
                "per": 10,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tasmin_per=tasmin_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class TN90p(ClimateIndicator):
    """Fraction of days with min temperature > 90th percentile."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tn90p

    def compute(
        self,
        base_period_time_range=None,
        tasmin_per=None,
        **params,
    ):
        """Calculate fraction of days with min temperature > 90th percentile.

        Parameters
        ----------
        tasmin_per: xr.DataArray, optional
            Minimum temperature 90th percentile reference value.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tasmin_per`.
            This will be used only if `tasmin_per` is None.

        Returns
        -------
        xarray.DataArray
            Fraction of days with minimum temperature > 90th percentile".

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tx90p
        """
        percentiles = {
            "tasmin_per": {
                "variable": "tasmin",
                "per": 90,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tasmin_per=tasmin_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class TNn(ClimateIndicator):
    """Minimum minimum temperature (tasmin)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tn_min

    def compute(self, **params):
        """Calculate minimum daily minimum temperature.

        Returns
        -------
        xarray.DataArray
            Minimum daily minimum temperature.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tn_min
        """
        return self.compute_climate_indicator(params=params)


class TNx(ClimateIndicator):
    """Maximum minimum temperature (tasmin)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.tn_max

    def compute(self, **params):
        """Calculate maximum daily minimum temperature.

        Returns
        -------
        xarray.DataArray
            Maximum daily minimum temperature.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tn_max
        """
        return self.compute_climate_indicator(params=params)


class WD(ClimateIndicator):
    """Number of warm and dry days (tas, pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.warm_and_dry_days

    def compute(
        self,
        base_period_time_range=None,
        tas_per=None,
        pr_per=None,
        **params,
    ):
        """Calculate number of warm and dry days.

        Parameters
        ----------
        tas_per: xr.DataArray, optional
            Mean temperature 75th percentile reference value above which
            a day is considered as a warm day.
        pr_per: xr.DataArray, optional
            Precipitation 25th percentile reference value below which
            a day is considered as a dry day.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tas_per` and/or `pr_per`.
            This will be used only if `tas_per` and/or `pr_per` is None.

        Returns
        -------
        xarray.DataArray:
            Number of days where warm and dry conditions coincide.
            If temperature is above {tas_per} a day is considered as a
            warm day.
            If precipitation is below {pr_per} a day is considered as a
            dry day.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.warm_and_dry_days
        """
        percentiles = {
            "tas_per": {
                "variable": "tas",
                "per": 75,
                "method": "temperature",
            },
            "pr_per": {
                "variable": "pr",
                "per": 25,
                "method": "precipitation",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tas_per=tas_per,
            pr_per=pr_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class WSDI(ClimateIndicator):
    """Warm spell duration index (tasmax)."""

    def __init__(self):
        super().__init__()
        self.window = 6
        self.func = xc.atmos.warm_spell_duration_index

    def compute(
        self,
        window=None,
        base_period_time_range=None,
        tasmax_per=None,
        **params,
    ):
        """Calculate warm spell duration index.

        Parameters
        ----------
        window: int, optional
            Minimum number of days with temperature above `tasmax_per`
            to qualify as a warm spell (default: 6).
        tasmin_per: xr.DataArray, optional
            Maximum temperature 90th percentile reference value.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tasmax_per`.
            This will be used only if `tasmax_per` is None.

        Returns
        -------
        xarray.DataArray
            Number of days part of a 90th percentile warm spell.
            At least {window} consecutive days.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.warm_spell_duration_index
        """
        percentiles = {
            "tasmax_per": {
                "variable": "tasmax",
                "per": 90,
                "method": "temperature",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            window=window,
            tasmax_per=tasmax_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class WW(ClimateIndicator):
    """Number of warm and wet days (tas, pr)."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.warm_and_wet_days

    def compute(
        self,
        base_period_time_range=None,
        tas_per=None,
        pr_per=None,
        **params,
    ):
        """Calculate number of warm and wet days.

        Parameters
        ----------
        tas_per: xr.DataArray, optional
            Mean temperature 75th percentile reference value above which
            a day is considered as a warm day.
        pr_per: xr.DataArray, optional
            Precipitation 75th percentile reference value above which
            a day is considered as a wet day.
        base_period_time_range: list, optional
            List with left bound is start year string and right bound
            is end year string for calculating `tas_per` and/or `pr_per`.
            This will be used only if `tas_per` and/or `pr_per` is None.

        Returns
        -------
        xarray.DataArray:
            Number of days where warm and wet conditions coincide.
            If temperature is above {tas_per} a day is considered as a
            warm day.
            If precipitation is above {pr_per} a day is considered as a
            wet day.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.warm_and_wet_days
        """
        percentiles = {
            "tas_per": {
                "variable": "tas",
                "per": 75,
                "method": "temperature",
            },
            "pr_per": {
                "variable": "pr",
                "per": 75,
                "method": "precipitation",
            },
        }
        return self.compute_climate_indicator(
            params=params,
            tas_per=tas_per,
            pr_per=pr_per,
            percentiles=percentiles,
            base_period_time_range=base_period_time_range,
        )


class CSf(ClimateIndicator):
    """Number of cold spells (tas)."""

    def __init__(self):
        super().__init__()
        self.thresh = -10
        self.window = 3
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.cold_spell_frequency

    def compute(self, thresh=None, window=None, **params):
        """Calculate number of cold spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold temperature below which a day is considered
            as a cold day (default: -10 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature below thresh
            to qualify as a cold spell (default: 3).

        Returns
        -------
        xarray.DataArray
            Number of cold spells of at least {window} consecutive days
            with temperature below {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.cold_spell_frequency
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class CSx(ClimateIndicator):
    """Maximum length of cold spells (tas)."""

    def __init__(self):
        super().__init__()
        self.thresh = -10
        self.window = 1
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.cold_spell_max_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate maximum length of cold spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold temperature below which a day is considered
            as a cold day (default: -10 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature below thresh
            to qualify as a cold spell (default: 1).

        Returns
        -------
        xarray.DataArray
            Number of cold spells of at least {window} consecutive days
            with temperature below {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.cold_spell_max_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class CSn(ClimateIndicator):
    """Total number of days in cold spells (tas)."""

    def __init__(self):
        super().__init__()
        self.thresh = -10
        self.window = 3
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.cold_spell_total_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate total number of days in cold spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold temperature below which a day is considered
            as a cold day (default: -10 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature below thresh
            to qualify as a cold spell (default: 3).

        Returns
        -------
        xarray.DataArray
            Total number of days in cold spells of at least {window}
            consecutive days with temperature below {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.cold_spell_total_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class HSf(ClimateIndicator):
    """Number of hot spells (tasmax)."""

    def __init__(self):
        super().__init__()
        self.thresh = 35
        self.window = 3
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.hot_spell_frequency

    def compute(self, thresh=None, window=None, **params):
        """Calculate number of hot spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold maximum temperature above which a day is considered
            as a hot day (default: 35 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above thresh
            to qualify as a hot spell (default: 3).

        Returns
        -------
        xarray.DataArray
            Number of hot spells of at least {window} consecutive days
            with maximum temperature above {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.hot_spell_frequency
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class HSx(ClimateIndicator):
    """Maximum lenght of hot spells (tasmax)."""

    def __init__(self):
        super().__init__()
        self.thresh = 35
        self.window = 1
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.hot_spell_max_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate maximum lenght of hot spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold maximum temperature above which a day is considered
            as a hot day (default: 35 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above thresh
            to qualify as a hot spell (default: 1).

        Returns
        -------
        xarray.DataArray
            Maximum length of hot spells of at least {window}
            consecutive days with maximum temperature above {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.hot_spell_max_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class HSn(ClimateIndicator):
    """Total number of days in hot spells (tasmax)."""

    def __init__(self):
        super().__init__()
        self.thresh = 35
        self.window = 3
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.hot_spell_total_length

    def compute(self, thresh=None, window=None, **params):
        """Calculate total number of days in hot spells.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold maximum temperature above which a day is considered
            as a hot day (default: 35 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above thresh
            to qualify as a hot spell (default: 3).

        Returns
        -------
        xarray.DataArray
            Total number of  days in hot spells of at least {window}
            consecutive days with maximum temperature above {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.hot_spell_total_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class SD(ClimateIndicator):
    """Number of snow days."""

    def __init__(self):
        super().__init__()
        self.low = 1
        self.units = {"low": "mm/day"}
        self.func = xc.atmos.days_with_snow

    def compute(self, low=None, **params):
        """Calculate number of snow days.

        Parameters
        ----------
        thresh: int or string
            Liquid water equivalent snowfall rate above which a day is
            considered as a snow day (default: 1 mm/day) .
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Number of days with solid precipitation flux
            above {thresh} threshold.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.days_with_snow
        """
        return self.compute_climate_indicator(params=params, low=low)


class SCD(ClimateIndicator):
    """Snow cover duration."""

    def __init__(self):
        super().__init__()
        self.thresh = 3
        self.units = {"thresh": "cm"}
        self.func = xc.land.snd_season_length

    def compute(
        self,
        thresh=None,
        **params,
    ):
        """Calculate snow cover duration.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold snow thickness above which a day is considered
            as a snow day (default: 3 cm).
            If type of threshold is an integer the unit is set to cm.

        Returns
        -------
        xarray.DataArray
            Number of days with snow cover above {thresh} threshold.
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class Sint(ClimateIndicator):
    """Snowfall intensity."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.snowfall_intensity

    def compute(self, thresh=None, **params):
        """Calculate snowfall intensity.

        Parameters
        ----------
        thresh: int or string, optional
            Liquid water equivalent snowfall rate above which a day is
            considered as a snow day (default: 1 mm/day) .
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Mean daily snowfall during days with snowfall > {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.snowfall_intensity
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class Sfreq(ClimateIndicator):
    """Snowfall frequency."""

    def __init__(self):
        super().__init__()
        self.thresh = 1
        self.units = {"thresh": "mm/day"}
        self.func = xc.atmos.snowfall_frequency

    def compute(self, thresh=None, **params):
        """Calculate snowfall frequency.

        Parameters
        ----------
        thresh: int or string, optional
            Liquid water equivalent snowfall rate above which a day is
            considered as a snow day (default: 1 mm/day) .
            If type of threshold is an integer the unit is set to mm/day.

        Returns
        -------
        xarray.DataArray
            Percentage of days with snowfall > {thresh}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.snowfall_frequency
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class UTCI(ClimateIndicator):
    """Universal thermal climate index."""

    def __init__(self):
        super().__init__()
        self.stat = "sunlit"
        self.mask_invalid = True
        self.func = xc.indicators.convert.universal_thermal_climate_index

    def compute(self, stat=None, mask_invalid=None, **params):
        """Calculate universal thermal climate index.

        Parameters
        ----------
        stat: {‘average’, ‘sunlit’, ‘instant’}, optional
            Which statistic to apply. If “average”, the average of the cosine
            of the solar zenith angle is calculated.
            If “instant”, the instantaneous cosine of the solar zenith angle is
            calculated.
            If “sunlit”, the cosine of the solar zenith angle is calculated
            during the sunlit period of each interval.
            If “instant”, the instantaneous cosine of the solar zenith angle is
            calculated.
            This is necessary if mrt is not None (default: average).
        mask_valid: bool, optional
            If True, UTCI values are NaN where any of the inputs are outside
            their validity ranges:
            -50°C < tas < 50°C,
            -30°C < tas - mrt < 30°C
            0.5 m/s < sfcWind < 17.0 m/s
            (default: True).

        Returns
        -------
        xarray.DataArray
            Universal Thermal Climate Index.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.universal_thermal_climate_index
        """
        return self.compute_climate_indicator(
            params=params, stat=stat, mask_invalid=mask_invalid
        )


class WI(ClimateIndicator):
    """Number of winter days (tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh = -10
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.tn_days_below

    def compute(self, thresh=None, **params):
        """Calculate number of winter days.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold minimum temperature below which a day is considered
            as a winter day (default: -10 degC).
            If type of threshold is an integer the unit is set to degC.

        Returns
        -------
        xarray.DataArray
            Number of winter days.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.tn_days_below
        """
        return self.compute_climate_indicator(params=params, thresh=thresh)


class HWf(ClimateIndicator):
    """Number of heat waves (tasmax, tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh_tasmin = 22
        self.thresh_tasmax = 30
        self.window = 3
        self.units = {"thresh_tasmin": "degC", "thresh_tasmax": "degC"}
        self.func = xc.atmos.heat_wave_frequency

    def compute(
        self,
        thresh_tasmin=None,
        thresh_tasmax=None,
        window=None,
        **params,
    ):
        """Calculate number of heat waves.

        Parameters
        ----------
        thresh_tasmin: int or string, optional
            Threshold minimum temperature above which a day is considered
            as a heat day (default: 22 degC).
            If type of threshold is an integer the unit is set to degC.
        thresh_tasmax: int or string, optional
            Threshold maximum temperature above which a day is considered
            as a winter day (default: 30 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above thresh
            to qualify as a hot spell (default: 3).

        Returns
        -------
        xarray.DataArray
            Number of heat waves of at least {window} consecutive days
            with maximum temperature above {thresh_tasmax} and
            minimum temperature above {thresh_tasmin}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.hot_spell_frequency
        """
        return self.compute_climate_indicator(
            params=params,
            thresh_tasmin=thresh_tasmin,
            thresh_tasmax=thresh_tasmax,
            window=window,
        )


class HWx(ClimateIndicator):
    """Maximum length of heat waves (tasmax, tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh_tasmin = 22
        self.thresh_tasmax = 30
        self.window = 1
        self.units = {"thresh_tasmin": "degC", "thresh_tasmax": "degC"}
        self.func = xc.atmos.heat_wave_max_length

    def compute(
        self,
        thresh_tasmax=None,
        thresh_tasmin=None,
        window=None,
        **params,
    ):
        """Calculate maximum number of heat waves.

        Parameters
        ----------
        thresh_tasmin: int or string, optional
            Threshold minimum temperature above which a day is considered
            as a heat day (default: 22 degC).
            If type of threshold is an integer the unit is set to degC.
        thresh_tasmax: int or string, optional
            Threshold maximum temperature above which a day is considered
            as a winter day (default: 30 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above thresh
            to qualify as a hot spell (default: 1).

        Returns
        -------
        xarray.DataArray
            Maximum length of heat waves of at least {window} consecutive days
            with maximum temperature above {thresh_tasmax} and
            minimum temperature above {thresh_tasmin}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.heat_wave_max_length
        """
        return self.compute_climate_indicator(
            params=params,
            thresh_tasmin=thresh_tasmin,
            thresh_tasmax=thresh_tasmax,
            window=window,
        )


class HWn(ClimateIndicator):
    """Total number of days in heat waves (tasmax, tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh_tasmin = 22
        self.thresh_tasmax = 30
        self.window = 3
        self.units = {"thresh_tasmin": "degC", "thresh_tasmax": "degC"}
        self.func = xc.atmos.heat_wave_total_length

    def compute(
        self,
        thresh_tasmin=None,
        thresh_tasmax=None,
        window=None,
        **params,
    ):
        """Calculate total number of days in heat waves.

        Parameters
        ----------
        thresh_tasmin: int or string, optional
            Threshold minimum temperature above which a day is considered
            as a heat day (default: 22 degC).
            If type of threshold is an integer the unit is set to degC.
        thresh_tasmax: int or string, optional
            Threshold maximum temperature above which a day is considered
            as a winter day (default: 30 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above thresh
            to qualify as a hot spell (default: 3).

        Returns
        -------
        xarray.DataArray
            Total number of days in heat waves of at least {window}
            consecutive days with maximum temperature above {thresh_tasmax}
            and minimum temperature above {thresh_tasmin}.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.hot_spell_total_length
        """
        return self.compute_climate_indicator(
            params=params,
            thresh_tasmin=thresh_tasmin,
            thresh_tasmax=thresh_tasmax,
            window=window,
        )


class GSS(ClimateIndicator):
    """Growing season start (tas)."""

    def __init__(self):
        super().__init__()
        self.thresh = 5
        self.window = 5
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.growing_season_start

    def compute(self, thresh=None, window=None, mid_date=None, **params):
        """Calculate growing season start.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold temperature above which the growing season starts
            (default: 5 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above threshold
            needed for evaluation (default: 5).

        Returns
        -------
        xarray.DataArray
            Growing season start.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.growing_season_start
        """
        return self.compute_climate_indicator(
            params=params,
            thresh=thresh,
            window=window,
            mid_date=mid_date,
        )


class GSE(ClimateIndicator):
    """Growing season end (tas)."""

    def __init__(self):
        super().__init__()
        self.thresh = 5
        self.window = 5
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.growing_season_end

    def compute(self, thresh=None, window=None, mid_date=None, **params):
        """Calculate growing season end.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold temperature below which the growing season ends
            (default: 5 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature below threshold
            needed for evaluation (default: 5).

        Returns
        -------
        xarray.DataArray
            Growing season end.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.growing_season_end
        """
        return self.compute_climate_indicator(
            params=params,
            thresh=thresh,
            window=window,
            mid_date=mid_date,
        )


class GSL(ClimateIndicator):
    """Growing season length (tas)."""

    def __init__(self):
        super().__init__()
        self.thresh = 5
        self.window = 5
        self.mid_date = "07-01"
        self.freq = "YS"
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.growing_season_length

    def compute(self, thresh=None, window=None, mid_date=None, **params):
        """Calculate growing season length.

        Parameters
         ----------
         thresh: Threshold temperature on which to base evaluation.
             (default: 5 degC). If type of threshold is an integer
             the unit is set to degC.
         window: int, optional
             Minimum number of days with temperature above threshold
             to mark the beginning and end of growing season (default: 5).
         mid_date: (date (string, MM-DD)) – Date of the year
             after which to look for the end of the season.
             Should have the format ‘%m-%d’. Default : 07-01

         Returns
         -------
         xarray.DataArray
             Growing season length.

         Notes
         -----
         For more information on the input parameters see:
             https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.growing_season_length
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class FFS(ClimateIndicator):
    """Frost free season start (tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh = 0
        self.window = 5
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.frost_free_season_start

    def compute(self, thresh=None, window=None, **params):
        """Calculate frost free season start.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold minimum temperature above which the frost free season
            starts (default: 0 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature above threshold
            needed for evaluation (default: 5).

        Returns
        -------
        xarray.DataArray
            Frost free season start.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.frost_free_season_start
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class FFE(ClimateIndicator):
    """Frost free season end (tasmin)."""

    def __init__(self):
        super().__init__()
        self.thresh = 0
        self.window = 5
        self.units = {"thresh": "degC"}
        self.func = xc.atmos.frost_free_season_end

    def compute(self, thresh=None, window=None, **params):
        """Calculate frost free season end.

        Parameters
        ----------
        thresh: int or string, optional
            Threshold minimum temperature below which the frost free season
            ends (default: 0 degC).
            If type of threshold is an integer the unit is set to degC.
        window: int, optional
            Minimum number of days with temperature below threshold
            needed for evaluation (default: 5).

        Returns
        -------
        xarray.DataArray
            Frost free season end.

        Notes
        -----
        For more information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.frost_free_season_end
        """
        return self.compute_climate_indicator(
            params=params, thresh=thresh, window=window
        )


class FG(ClimateIndicator):
    """Mean daily mean wind speed."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.sfcWind_mean

    def compute(self, **params):
        """Calculate mean daily mean wind speed.

        Returns
        -------
        xarray.DataArray
            Mean daily mean wind speed.

        Notes
        -----
        If sfcWind is not provided in `ds` sfcWind could be computed
        from `ds`.uas and `ds`.vas.

        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.sfcWind_mean
        """
        return self.compute_climate_indicator(params=params)


class FGn(ClimateIndicator):
    """Minimum daily mean wind speed."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.sfcWind_min

    def compute(self, **params):
        """Calculate minimum daily mean wind speed.

        Returns
        -------
        xarray.DataArray
            Minimum daily mean wind speed.

        Notes
        -----
        If sfcWind is not provided in `ds` sfcWind could be computed
        from `ds`.uas and `ds`.vas.

        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.sfcWind_min
        """
        return self.compute_climate_indicator(params=params)


class FGx(ClimateIndicator):
    """Maximum daily mean wind speed."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.sfcWind_max

    def compute(self, **params):
        """Calculate maximum daily mean wind speed.

        Returns
        -------
        xarray.DataArray
            Maximum daily mean wind speed.

        Notes
        -----
        If sfcWind is not provided in `ds` sfcWind could be computed
        from `ds`.uas and `ds`.vas.

        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.sfcWind_max
        """
        return self.compute_climate_indicator(params=params)


class FX(ClimateIndicator):
    """Mean daily maximum wind speed."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.sfcWindmax_mean

    def compute(self, **params):
        """Calculate mean daily maximum wind speed.

        Returns
        -------
        xarray.DataArray
            Mean daily maximum wind speed.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.sfcWindmax_mean
        """
        return self.compute_climate_indicator(params=params)


class FXn(ClimateIndicator):
    """Minimum daily maximum wind speed."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.sfcWindmax_min

    def compute(self, **params):
        """Calculate minimum daily maximum wind speed.

        Returns
        -------
        xarray.DataArray
            Minimum daily maximum wind speed.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.sfcWindmax_min
        """
        return self.compute_climate_indicator(params=params)


class FXx(ClimateIndicator):
    """Maximum daily maximum wind speed."""

    def __init__(self):
        super().__init__()
        self.func = xc.atmos.sfcWindmax_max

    def compute(self, **params):
        """Calculate maximum daily maximum wind speed.

        Returns
        -------
        xarray.DataArray
            Maximum daily maximum wind speed.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.sfcWindmax_max
        """
        return self.compute_climate_indicator(params=params)


class HIX(ClimateIndicator):
    """temperature felt by a person.

    When relative humidity is taken into account (tas, hurs)

    """

    def __init__(self):
        super().__init__()
        self.func = xc.indicators.convert.humidex

    def compute(self, **params):
        """Calculate maximum number of consecutive heat days.

        Returns
        -------
        xarray.DataArray
            temperature felt by a person when relative humidity
            is taken into account.

        Notes
        -----
        For information on the input parameters see:
            https://xclim.readthedocs.io/en/stable/api.html#xclim.indicators.atmos.humidex
        """
        return self.compute_climate_indicator(params=params)



import os
import warnings
from pathlib import Path

from pyhomogenize import save_xrdataset

from ._tables import pjson
from ._utils import object_attrs_to_self


class OutputWriter:
    """Class for writing xarray.Dataset to disk.
    Get default project-specific output name.

    Parameters
    ----------
    postproc_obj: index_calculator.postprocessing or list
        ``index_calculator.postprocessing`` object
    output_name: bool or str (default: True), optional
        If True get project-specific output name.
        If `str` set output_name to that string.
    output_dir: str (default: "."), optional
        Set output directory
    drs: bool (default: True), optional
        If True create project-specific directory structure

    Example
    -------
    Write climate indicator dataset to disk::

        from pyhomogenize import open_xrdataset
        from index_calculator import preprocessing
        from index_calculator import processing
        from index_calculator import postprocessing
        from index_calculator import outputwriter

        netcdf_file = "tas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_"
                      "GERICS-REMO2015_v1_day_20010101-20051231.nc"
        ds = open_xrdataset(netcdf_file)

        preproc = preprocessing(ds)
        proc = processing(index="TG", preproc_obj=preproc)
        postproc = postprocessing(
            project="CORDEX",
            proc_obj=proc,
            institution="Helmholtz-Zentrum hereon GmbH,"
                        "Climate Service Center Germany",
            institution_id="GERICS",
            contact="gerics-cordex@hereon.de",
        )
        outputwriter(postproc_obj=postproc)


        --> File written: cordex/climdex/EUR-11/GERICS/GERICS/MPI-M-MPI-ESM-LR/
                          historical/r3i1p1/GERICS-REMO2015/v1/year/TG/
                          TG_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_
                          GERICS-REMO2015_v1_day_GERICS_year_2001-2005.nc

    """

    def __init__(
        self,
        postproc_obj=None,
        output_name=True,
        output_dir=".",
        drs=True,
        **kwargs,
    ):
        if postproc_obj is None:
            raise NameError(
                "Please select an index_calculator.PostProcessing object."
                "'proc_obj=...'"
            )
        object_attrs_to_self(postproc_obj, self)
        if not isinstance(self.postproc, list):
            postprocs = [self.postproc]
        else:
            postprocs = self.postproc
        if drs is True:
            output_dir = os.path.join(
                output_dir, self._directory_structure(postprocs[0])
            )
        os.makedirs(output_dir, exist_ok=True)
        self.output_name = output_name
        ds_dict = {}
        for postproc in postprocs:
            ds_ = postproc.copy()
            if self.output_name is True:
                output_name = self._outname(ds_)
            else:
                output_name = self.output_name
            if not output_name:
                print("Could not write output.")
                continue
            outputname = Path(os.path.join(output_dir, output_name)).resolve()
            ds_dict[output_name] = self._to_netcdf(ds_, outputname)

    def _parse_components_to_format(self, ds, out_components, out_format):
        def test_ocomp(ocomp):
            return ocomp.replace("/", "").replace(" ", "").replace(".", "-")

        ocomps = []
        for comp in out_components:
            if hasattr(ds, comp):
                ocomps.append(test_ocomp(getattr(ds, comp)))
            elif hasattr(self.proc, comp):
                ocomps.append(test_ocomp(getattr(self.proc, comp)))
            elif hasattr(self.ds, comp):
                ocomps.append(test_ocomp(getattr(self.ds, comp)))
            else:
                ocomps.append("NA")
                warnings.warn(f"{comp} not found!")
        return out_format.format(*ocomps)

    def _outname(self, ds):
        """Set project-specific ouput file name."""

        try:
            output_fmt = pjson[self.project]["format"]
            output_comps = pjson[self.project]["components"]
        except ValueError:
            warnings.warn(f"Project {self.project} not known")
            return
        return self._parse_components_to_format(ds, output_comps, output_fmt)

    def _directory_structure(self, ds):
        """Set project-specific directory structure."""
        try:
            output_fmt = pjson[self.project]["drs_format"]
            output_comps = pjson[self.project]["drs_components"]
        except ValueError:
            warnings.warn(f"Project {self.project} not known")
            return
        return self._parse_components_to_format(ds, output_comps, output_fmt)

    def _to_netcdf(self, ds, name):
        """Write xarray.Dataset to netCDF file on disk."""
        ds = save_xrdataset(
            ds,
            name=name,
            encoding_dict={
                "encoding": {
                    self.CIname: {"dtype": "float32"},
                }
            },
            unlimited_dims={self.unlimited_dims: True},
        )
        print(f"File written: {name}")
        return ds



"""Top-level package for index_calculator."""

__author__ = """Ludwig Lierhammer"""
__email__ = "ludwig.lierhammer@dwd.de"

from ._outputwriter import OutputWriter as outputwriter
from ._postprocessing import PostProcessing as postprocessing
from ._preprocessing import PreProcessing as preprocessing
from ._processing import Processing as processing
from ._tables import (
    cfjson,  # noqa
    fjson,  # noqa
    mjson,  # noqa
    pjson,  # noqa
    vjson,  # noqa
    xjson,  # noqa
)
from .index_calculator import IndexCalculator as index_calculator


def _get_version():
    __version__ = "unknown"
    try:
        from ._version import __version__
    except ImportError:
        pass
    return __version__


__version__ = _get_version()

preprocessing.__module__ = __name__
preprocessing.__name__ = "preprocessing"
processing.__module__ = __name__
processing.__name__ = "processing"
postprocessing.__module__ = __name__
postprocessing.__name__ = "postprocessing"
outputwriter.__module__ = __name__
outputwriter.__name__ = "outputwriter"
index_calculator.__module__ = __name__
index_calculator.__name__ = "index_calculator"



import importlib.resources as resources
import json


def read_from_json(jsfile):
    """Read json file from package resources.

    Parameters
    ----------
    jsfile: str
        Json file path relative to the package.

    Returns
    -------
    dict
       Python dictionary.
    """
    with (
        resources.files("index_calculator")
        .joinpath(jsfile)
        .open("r", encoding="utf-8") as f
    ):
        return json.load(f)


mjson = read_from_json("tables/metadata.json")
xjson = read_from_json("tables/xcalc.json")
pjson = read_from_json("tables/projects.json")
vjson = read_from_json("tables/input_vars.json")
cfjson = read_from_json("tables/cf_conversion.json")
fjson = read_from_json("tables/convert_to_frequency.json")



"""Console script for index_calculator."""

import argparse
import sys

import dask  # noqa
from dask.distributed import Client
from pyhomogenize import open_xrdataset

import index_calculator as xcalc


def _parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input_file",
        dest="input_file",
        nargs="?",
        help="netCDF input file name",
    )
    parser.add_argument(
        "-x",
        "--index",
        dest="climate_index",
        nargs="?",
        help="climate index short name",
    )
    parser.add_argument(
        "-freq",
        "--frequency",
        dest="frequency",
        default="year",
        help="output file frequency",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project",
        default="N/A",
        help="MIP project name",
    )
    parser.add_argument(
        "-inst",
        "--institute",
        dest="institution",
        default="N/A",
        help="institution long name",
    )
    parser.add_argument(
        "-inst_id",
        "--institute_id",
        dest="institution_id",
        default="N/A",
        help="institution id",
    )
    parser.add_argument(
        "-contact",
        "--contact",
        dest="contact",
        default="N/A",
        help="institution mail contact",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        default=True,
        help="netCDF output file name. Necessary if project is not selected.",
    )
    return parser


def _args_to_xcalc(args):
    ds = open_xrdataset(args.input_file)
    return xcalc.index_calculator(
        ds=ds,
        index=args.climate_index,
        freq=args.frequency,
        project=args.project,
        institution=args.institution,
        institution_id=args.institution_id,
        contact=args.contact,
        output=args.output,
    ).compute(write=True)


def main():
    parser = _parser()
    args = parser.parse_args()

    _args_to_xcalc(args)
    return 0


if __name__ == "__main__":
    with Client() as client:
        sys.exit(main())  # pragma: no cover



import pyhomogenize as pyh
import xarray as xr
from pyhomogenize._consts import fmt as _fmt

from ._consts import _bounds
from ._tables import cfjson, fjson
from ._utils import check_existance, get_time_range_as_str, kwargs_to_self


class PreProcessing:
    """Class for pre-processing xarray datasets.

    Parameters
    ----------
    ds : xr.Dataset
        xarray Dataset.
    project: {"CORDEX", "CMIP5", "CMIP6", "EOBS", "ERA5", "N/A"}
        (default: "N/A), optional
        Project name
    var_name : str or list, optional
        CF variable(s) contained in `ds`.
        If None (default) `var_name` is read from `ds` with pyhomogenize.
    freq: str (default="year"), optional
        Climate indicator output frequency
    ifreq: str (default="day"), optional
        Climate indicator input frequency
    time_range: list, optional
        List of two strings representing the left and right time bounds.
        Select time slice with those limits from `ds`.
    crop_time_axis: bool, optional
        If True (default) select time slice from `ds`.
        The left and the right bounds depends on `freq`.
        For example: If `freq` is year the left bound has to be January, 1st
        and the right bound has to be the last day of December.
    check_time_axis: bool, optional
        If True (default) check the time axis on duplicated, redundant
        and/or missing time steps.

    Example
    -------
    Do some preprocessing with a netcdf file on disk::

        from pyhomogenize import open_xrdataset
        from index_calculator import preprocessing

        netcdf_file = "tas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_"
                      "GERICS-REMO2015_v1_day_20010101-20051231.nc"
        ds = open_xrdataset(netcdf_file)

        preproc = preprocessing(ds)

        preproc_ds = preproc.preproc
    """

    def __init__(
        self,
        ds=None,
        project="N/A",
        var_name=None,
        freq="year",
        ifreq="day",
        time_range=None,
        crop_time_axis=True,
        check_time_axis=True,
        **kwargs,
    ):
        if ds is None:
            raise ValueError("Please select an input xarray dataset. 'ds=...'")

        self.ds = ds
        self.project = check_existance({"project": project}, self)
        self.var_name = var_name
        self.freq = freq
        self.ifreq = ifreq
        self.fmt = _fmt[freq].replace("-", "")
        self.afmt = _fmt[ifreq].replace("-", "")
        self.time_range = time_range
        self.crop_time_axis = crop_time_axis
        self.check_time_axis = check_time_axis
        kwargs_to_self(kwargs, self)
        self.preproc = self._preprocessing()

    def _rename_variable_names(self, ds):
        if self.project not in cfjson.keys():
            return ds
        var_names = cfjson[self.project]["variables"]
        units = cfjson[self.project]["units"]
        for dvar in ds.data_vars:
            if dvar in var_names.keys():
                ds = ds.rename({dvar: var_names[dvar]})
                dvar = var_names[dvar]
            if dvar in units.keys():
                ds[dvar].attrs["units"] = units[dvar]
        return ds

    def _convert_to_frequency(self, ds):
        if self.ifreq not in fjson.keys():
            raise ValueError(
                f"Could not convert to frequency {self.ifreq}",
                f"Try one of {fjson.keys()}.",
            )
        conv = fjson[self.ifreq]
        if conv["freq"] == xr.infer_freq(ds.time):
            return ds
        data_vars = {}
        for dvar in ds.data_vars:
            if dvar in conv["var"].keys():
                data_vars[dvar] = getattr(
                    ds[dvar].resample(time=conv["freq"]),
                    conv["var"][dvar],
                )(dim="time")
                data_vars[dvar].attrs["cell_methods"] = "time: {}".format(
                    conv["var"][dvar]
                )
                coords = data_vars[dvar].coords
        return xr.Dataset(
            data_vars=data_vars,
            coords=coords,
            attrs=ds.attrs,
        )

    def _preprocessing(self):
        ds_ = self._convert_to_frequency(self.ds)
        time_control = pyh.time_control(ds_)
        if not self.var_name:
            self.var_name = time_control.name

        avail_time = get_time_range_as_str(time_control.time, self.afmt)

        if self.time_range:
            time_control.select_time_range(self.time_range)
        if self.crop_time_axis:
            time_control.select_limited_time_range(
                smonth=_bounds[self.freq]["start"],
                emonth=_bounds[self.freq]["end"],
            )
        if self.check_time_axis:
            time_control.check_timestamps(correct=True)
        self.ATimeRange = avail_time
        ds = time_control.ds
        return self._rename_variable_names(ds)



import cf_xarray  # noqa
import cftime  # noqa
import numpy as np
import pyhomogenize as pyh
import xarray as xr
from pyhomogenize._consts import freqs as _freq
from pyhomogenize._consts import frequencies as _tfreq

from . import _indices as indices
from ._tables import vjson
from ._utils import (
    check_existance,
    get_alpha_name,
    get_numb_name,
    get_replace_name,
    kwargs_to_self,
    object_attrs_to_self,
)
from ._variable_conversion import ConvertVariables


class Processing:
    """Class for processing ``index_calculator.preprocessing`` object.

    Parameters
    ----------
    index: str
        Climate indicator name to be calculated.
    preproc_obj: index_calculator.preprocessing
        ``index_calculator.preprocessing`` object

    Example
    -------
    Calculate a climate indicator `TG` from netcdf file on disk::

        from pyhomogenize import open_xrdataset
        from index_calculator import preprocessing
        from index_calculator import processing

        netcdf_file = "tas_EUR-11_MPI-M-MPI-ESM-LR_historical_r3i1p1_"
                      "GERICS-REMO2015_v1_day_20010101-20051231.nc"
        ds = open_xrdataset(netcdf_file)

        preproc = preprocessing(ds)
        proc = processing(index="TG", preproc_obj=preproc)

        proc_ds = proc.proc
    """

    def __init__(
        self,
        index=None,
        preproc_obj=None,
        **kwargs,
    ):
        if preproc_obj is None:
            raise ValueError(
                "Please select an index_calculator.PreProcessing object."
                "'preproc_obj='...'"
            )
        object_attrs_to_self(preproc_obj, self)
        self.CIname = check_existance({"index": index}, self)
        kwargs_to_self(kwargs, self)
        self._get_idx_name_and_repl()
        self.proc = self._processing()

    def _get_numb_name_and_idx_object(self):
        alpha_name = get_alpha_name(self.CIname)
        numb_name = get_numb_name(self.CIname)
        replace_name = get_replace_name(self.CIname)
        if hasattr(indices, self.CIname):
            idx_object = getattr(indices, self.CIname)
            self.IDXname = self.CIname
            numb_name = ""
        elif hasattr(indices, alpha_name):
            idx_object = getattr(indices, alpha_name)
            self.IDXname = alpha_name
        elif hasattr(indices, replace_name):
            idx_object = getattr(indices, replace_name)
            self.IDXname = replace_name
        else:
            raise NameError(f"{self.CIname} not defined.")
        return numb_name, idx_object()

    def _get_replacement(self, obj, numb_name):
        replacement = {}
        for attr in dir(obj):
            if attr[0] == "_":
                continue
            if callable(getattr(obj, attr)):
                continue
            if attr in self.kwargs.keys():
                replacement[attr] = self.kwargs[attr]
            elif numb_name:
                default_value = getattr(obj, attr)
                if isinstance(default_value, (int, str, float)):
                    default_value = float(default_value)
                    if default_value < 0:
                        numb_name = f"-{numb_name}"
                else:
                    continue
                replacement[attr] = float(numb_name)
            else:
                replacement[attr] = getattr(obj, attr)
            if isinstance(replacement[attr], list):
                continue
            if isinstance(replacement[attr], str):
                if not replacement[attr].isnumeric():
                    continue
                if replacement[attr][0] == "0":
                    replacement[attr] = float(
                        f"{replacement[attr][0]}.{replacement[attr][1:]}"
                    )
                else:
                    replacement[attr] = int(replacement[attr])

        return replacement

    def _get_idx_name_and_repl(self):
        numb_name, idx_object = self._get_numb_name_and_idx_object()
        defaults = [attr for attr in dir(idx_object)]
        object_attrs_to_self(idx_object, self, overwrite=False)
        self.replacement = self._get_replacement(
            idx_object,
            numb_name,
        )
        for k, v in self.replacement.items():
            if k in defaults and k in self.kwargs.keys():
                if "YY" in self.CIname:
                    self.CIname = self.CIname.replace("YY", str(v))
                elif isinstance(v, list):
                    continue
                elif isinstance(v, bool):
                    continue
                elif isinstance(v, str):
                    continue
                elif numb_name:
                    self.CIname = self.CIname.replace(
                        numb_name,
                        str(v),
                    )
                elif str(v) not in self.CIname:
                    self.CIname = f"{self.CIname}{str(v)}"
                break

    def _adjust_params_to_ci(self):
        params = {
            "ds": self.preproc,
            "freq": _freq[self.freq],
        }
        params.update(self.replacement)
        return params

    def _processing(self):
        """Calculate climate index."""
        ds = self.preproc
        conv_vars = ConvertVariables(ds, **self.kwargs)
        input_variables = vjson[self.IDXname]
        if isinstance(input_variables, str):
            input_variables = [input_variables]
        for input_variable in input_variables:
            if input_variable in ds.data_vars:
                continue
            if not hasattr(conv_vars, input_variable):
                raise ValueError(
                    f"""Requested input variable {input_variable}
                    is not provided and can not be calculated
                    from provided input variables {ds.data_vars}"""
                )
            ds[input_variable] = getattr(conv_vars, input_variable)()

        dvars = ds.data_vars
        data_vars = {
            k: v
            for k, v in dvars.items()
            if k not in self.var_name and "time" not in ds[k].coords
        }
        params = self._adjust_params_to_ci()
        array = self.compute(**params)
        basics = pyh.basics()
        for data_var in data_vars:
            if "grid_mapping" in ds[data_var].attrs:
                gm = ds[data_var].attrs["grid_mapping"]
                array.attrs["grid_mapping"] = gm
                break
        coords = {k: v for k, v in ds.coords.items() if "time" not in k}
        for k, v in coords.items():
            array[k] = v

        data_vars[self.CIname] = array
        idx_ds = xr.Dataset(
            data_vars=data_vars,
            coords=coords,
            attrs=self.preproc.attrs,
        )
        if "time" in array.coords:
            date_range = basics.date_range(
                start=self.preproc.time.values[0],
                end=self.preproc.time.values[-1],
                frequency=_tfreq[self.freq],
            )
            if len(idx_ds["time"]) == len(date_range):
                idx_ds = idx_ds.assign_coords(
                    {"time": date_range},
                )
            else:
                self.freq = self.ifreq
            dim_squeeze = []
            for dim in idx_ds.dims:
                if len(idx_ds[dim]) == 1 and dim != "time":
                    dim_squeeze += [dim]
            idx_ds = idx_ds.squeeze(dim=dim_squeeze)
            time_encoding = ds.time.encoding
            time_encoding["dtype"] = np.float64
            idx_ds.time.encoding = time_encoding
            idx_ds = (
                pyh.time_control(idx_ds)
                .add_time_bounds(
                    frequency=self.freq,
                )
                .ds
            )
            self.unlimited_dims = "time"
        else:
            for dim in idx_ds[self.CIname].dims:
                if dim not in self.preproc[self.var_name[0]].dims:
                    self.unlimited_dims = dim
                    break
            idx_ds[self.unlimited_dims] = idx_ds[self.unlimited_dims].astype(
                float,
            )
            transpose_list = list(
                map(
                    lambda x: x.replace("time", self.unlimited_dims),
                    self.preproc[self.var_name[0]].dims,
                ),
            )
            idx_ds[self.CIname] = idx_ds[self.CIname].transpose(
                *transpose_list,
            )
        for data_var in idx_ds.data_vars:
            data_var_repl = data_var.replace("bounds", "bnds")
            idx_ds = idx_ds.rename({data_var: data_var_repl})
        return idx_ds



import os

import pytest  # noqa

from .conftest import tas_day_netcdf


def test_cli():
    s = (
        "index_calculator "
        f"-i {tas_day_netcdf()} "
        "-o test.nc "
        "-x TG "
        "-p CORDEX "
        "-inst test_institution "
        "-inst_id TEST "
        "-contact test@test.de "
        "-freq week"
    )
    os.system(s)
